@@ -6,103 +6,120 @@
 
 module HaddockHtml ( 
 	ppHtml, copyHtmlBits, 
-	ppHtmlIndex, ppHtmlContents
+	ppHtmlIndex, ppHtmlContents,
+	ppHtmlHelpFiles
   ) where
 
 import Prelude hiding (div)
-import HaddockVersion
+
+import Binary ( openBinaryFile )
+import HaddockDevHelp
+import HaddockHH
+import HaddockHH2
+import HaddockModuleTree
 import HaddockTypes
 import HaddockUtil
-import HaddockModuleTree
-import HaddockHH
+import HaddockVersion
 import HsSyn
-
-import IO
-import Maybe	( fromJust, isJust )
-import List 	( sortBy )
-import Char	( isUpper, toUpper, isAlpha, ord )
-import Monad	( when, unless )
-
-#if __GLASGOW_HASKELL__ < 503
-import FiniteMap
-import URI	( escapeString, unreserved )
-#else
-import Data.FiniteMap
-import Network.URI ( escapeString, unreserved )
-#endif
-
 import Html
 import qualified Html
+import Map ( Map )
+import qualified Map hiding ( Map )
 
--- -----------------------------------------------------------------------------
--- Files we need to copy from our $libdir
-
-cssFile, iconFile :: String
-cssFile  = "haddock.css"
-iconFile = "haskell_icon.gif"
+import Control.Exception ( bracket )
+import Control.Monad ( when, unless )
+import Data.Char ( isUpper, toUpper )
+import Data.List ( sortBy )
+import Data.Maybe ( fromJust, isJust, mapMaybe )
+import Foreign.Marshal.Alloc ( allocaBytes )
+import System.IO ( IOMode(..), hClose, hGetBuf, hPutBuf )
 
 -- -----------------------------------------------------------------------------
 -- Generating HTML documentation
 
 ppHtml	:: String
+	-> Maybe String				-- package
 	-> Maybe String
-	-> [(Module, Interface)]
+	-> [Interface]
 	-> FilePath			-- destination directory
 	-> Maybe Doc			-- prologue text, maybe
-	-> Bool				-- do MS Help stuff
+	-> Maybe String		-- the Html Help format (--html-help)
 	-> Maybe String			-- the contents URL (--use-contents)
 	-> Maybe String			-- the index URL (--use-index)
 	-> IO ()
 
-ppHtml doctitle source_url ifaces odir prologue do_ms_help 
+ppHtml doctitle maybe_package source_url ifaces odir prologue maybe_html_help_format
 	maybe_contents_url maybe_index_url =  do
   let
 	visible_ifaces = filter visible ifaces
-	visible (_, i) = OptHide `notElem` iface_options i
+	visible i = OptHide `notElem` iface_options i
 
   when (not (isJust maybe_contents_url)) $ 
-    ppHtmlContents odir doctitle maybe_index_url 
-	[ (m,iface{iface_package=Nothing}) | (m,iface) <- visible_ifaces ]
+    ppHtmlContents odir doctitle maybe_package maybe_html_help_format maybe_index_url 
+	[ iface{iface_package=Nothing} | iface <- visible_ifaces ]
 	-- we don't want to display the packages in a single-package contents
 	prologue
 
   when (not (isJust maybe_index_url)) $ 
-    ppHtmlIndex odir doctitle maybe_contents_url visible_ifaces
-
-  -- Generate index and contents page for MS help if requested
-  when do_ms_help $ do
-    ppHHContents odir (map fst visible_ifaces)
-    ppHHIndex odir visible_ifaces
+    ppHtmlIndex odir doctitle maybe_package maybe_html_help_format maybe_contents_url visible_ifaces
+    
+  when (not (isJust maybe_contents_url && isJust maybe_index_url)) $ 
+	ppHtmlHelpFiles doctitle maybe_package ifaces odir maybe_html_help_format []
 
   mapM_ (ppHtmlModule odir doctitle source_url 
 	   maybe_contents_url maybe_index_url) visible_ifaces
 
+ppHtmlHelpFiles	
+    :: String                   -- doctitle
+    -> Maybe String				-- package
+	-> [Interface]
+	-> FilePath                 -- destination directory
+	-> Maybe String             -- the Html Help format (--html-help)
+	-> [FilePath]               -- external packages paths
+	-> IO ()
+ppHtmlHelpFiles doctitle maybe_package ifaces odir maybe_html_help_format pkg_paths =  do
+  let
+	visible_ifaces = filter visible ifaces
+	visible i = OptHide `notElem` iface_options i
+
+  -- Generate index and contents page for Html Help if requested
+  case maybe_html_help_format of
+    Nothing        -> return ()
+    Just "mshelp"  -> ppHHProject odir doctitle maybe_package visible_ifaces pkg_paths
+    Just "mshelp2" -> do
+		ppHH2Files      odir maybe_package visible_ifaces pkg_paths
+		ppHH2Collection odir doctitle maybe_package
+    Just "devhelp" -> ppDevHelpFile odir doctitle maybe_package visible_ifaces
+    Just format    -> fail ("The "++format++" format is not implemented")
+
+copyFile :: FilePath -> FilePath -> IO ()
+copyFile fromFPath toFPath =
+	(bracket (openBinaryFile fromFPath ReadMode) hClose $ \hFrom ->
+	 bracket (openBinaryFile toFPath WriteMode) hClose $ \hTo ->
+	 allocaBytes bufferSize $ \buffer ->
+		copyContents hFrom hTo buffer)
+	where
+		bufferSize = 1024
+		
+		copyContents hFrom hTo buffer = do
+			count <- hGetBuf hFrom buffer bufferSize
+			when (count > 0) $ do
+				hPutBuf hTo buffer count
+				copyContents hFrom hTo buffer
+
 
 copyHtmlBits :: FilePath -> FilePath -> Maybe FilePath -> IO ()
 copyHtmlBits odir libdir maybe_css = do
   let 
 	css_file = case maybe_css of
-			Nothing -> libdir ++ pathSeparator:cssFile
+			Nothing -> pathJoin [libdir, cssFile]
 			Just f  -> f
-	css_destination = odir ++ pathSeparator:cssFile
-
-	icon_file        = libdir ++ pathSeparator:iconFile
-	icon_destination = odir   ++ pathSeparator:iconFile
-  
-  css_contents <- readFile css_file
-  writeFile css_destination css_contents
-  icon_contents <- readFile icon_file
-  writeFile icon_destination icon_contents
-
-
-contentsHtmlFile, indexHtmlFile :: String
-contentsHtmlFile = "index.html"
-indexHtmlFile = "doc-index.html"
-
-subIndexHtmlFile :: Char -> String
-subIndexHtmlFile a = "doc-index-" ++ b ++ ".html"
-   where b | isAlpha a = [a]
-           | otherwise = show (ord a)
+	css_destination = pathJoin [odir, cssFile]
+	copyLibFile f = do
+	   copyFile (pathJoin [libdir, f]) (pathJoin [odir, f])
+ 
+  copyFile css_file css_destination
+  mapM_ copyLibFile [ iconFile, plusFile, minusFile, jsFile ]
 
 footer :: HtmlTable
 footer = 
@@ -113,15 +130,25 @@ footer =
 	)
    
 
-src_button :: Maybe String -> String -> String -> HtmlTable
-src_button source_url _ file
+src_button :: Maybe String -> String -> Interface -> HtmlTable
+src_button source_url _ iface
   | Just u <- source_url = 
-	let src_url = if (last u == '/') then u ++ file else u ++ '/':file
+	let src_url = spliceSrcURL iface u
 	in
 	topButBox (anchor ! [href src_url] << toHtml "Source code")
   | otherwise =
 	Html.emptyTable
   
+spliceSrcURL :: Interface -> String -> String
+spliceSrcURL iface url = run url
+  where run "" = ""
+        run ('%':'M':rest) = modl_str ++ run rest
+        run ('%':'F':rest) = iface_filename iface ++ run rest
+	run (c:rest) = c : run rest
+	
+	modl_str = case iface_module iface of { Module m -> 
+		   map (\x -> if x == '.' then '/' else x) m }
+
 contentsButton :: Maybe String -> HtmlTable
 contentsButton maybe_contents_url 
   = topButBox (anchor ! [href url] << toHtml "Contents")
@@ -156,7 +183,7 @@ pageHeader mdl iface doctitle source_url maybe_contents_url maybe_index_url =
   	image ! [src "haskell_icon.gif", width "16", height 16, alt " "]
        ) <->
        (tda [theclass "title"] << toHtml doctitle) <->
-	src_button source_url mdl (iface_filename iface) <->
+	src_button source_url mdl iface <->
 	contentsButton maybe_contents_url <->
 	indexButton maybe_index_url
     )
@@ -170,33 +197,50 @@ pageHeader mdl iface doctitle source_url maybe_contents_url maybe_index_url =
 
 moduleInfo :: Interface -> HtmlTable
 moduleInfo iface = 
-  case iface_info iface of
-    Nothing   -> Html.emptyTable
-    Just info ->
-          tda [align "right"] << narrowTable << (
-        	  (tda [theclass "infohead"] << toHtml "Portability") <->
-        	  (tda [theclass "infoval"] << toHtml (portability info)) </>
-        	  (tda [theclass "infohead"] << toHtml "Stability") <->
-        	  (tda [theclass "infoval"] << toHtml (stability info)) </>
-        	  (tda [theclass "infohead"] << toHtml "Maintainer") <->
-        	  (tda [theclass "infoval"] << toHtml (maintainer info))
-              )
+   let
+      info = iface_info iface
+
+      doOneEntry :: (String,ModuleInfo -> Maybe String) -> Maybe HtmlTable
+      doOneEntry (fieldName,field) = case field info of
+         Nothing -> Nothing
+         Just fieldValue -> 
+            Just ((tda [theclass "infohead"] << toHtml fieldName)
+               <-> (tda [theclass "infoval"]) << toHtml fieldValue)
+     
+      entries :: [HtmlTable]
+      entries = mapMaybe doOneEntry [
+         ("Portability",portability),
+         ("Stability",stability),
+         ("Maintainer",maintainer)
+         ]
+   in
+      case entries of
+         [] -> Html.emptyTable
+         _ -> tda [align "right"] << narrowTable << (foldl1 (</>) entries)
 
 -- ---------------------------------------------------------------------------
 -- Generate the module contents
 
 ppHtmlContents
-   :: FilePath -> String
+   :: FilePath
+   -> String
    -> Maybe String
-   -> [(Module,Interface)] -> Maybe Doc
+   -> Maybe String
+   -> Maybe String
+   -> [Interface] -> Maybe Doc
    -> IO ()
-ppHtmlContents odir doctitle maybe_index_url
+ppHtmlContents odir doctitle maybe_package maybe_html_help_format maybe_index_url
   mdls prologue = do
-  let tree = mkModuleTree [(mod,iface_package iface) | (mod,iface) <- mdls]
+  let tree = mkModuleTree 
+         [(iface_module iface,
+	   iface_package iface,
+	   toDescription iface) | iface <- mdls]
       html = 
-	header (thetitle (toHtml doctitle) +++
-		thelink ! [href cssFile, 
-		  rel "stylesheet", thetype "text/css"]) +++
+	header 
+		(documentCharacterEncoding +++
+		 thetitle (toHtml doctitle) +++
+		 styleSheet +++
+		 (script ! [src jsFile, thetype "text/javascript"] $ noHtml)) +++
         body << vanillaTable << (
    	    simpleHeader doctitle Nothing maybe_index_url </>
 	    ppPrologue doctitle prologue </>
@@ -204,7 +248,15 @@ ppHtmlContents odir doctitle maybe_index_url
 	    s15 </>
 	    footer
 	  )
-  writeFile (odir ++ pathSeparator:contentsHtmlFile) (renderHtml html)
+  writeFile (pathJoin [odir, contentsHtmlFile]) (renderHtml html)
+  
+  -- Generate contents page for Html Help if requested
+  case maybe_html_help_format of
+    Nothing        -> return ()
+    Just "mshelp"  -> ppHHContents  odir doctitle maybe_package tree
+    Just "mshelp2" -> ppHH2Contents odir doctitle maybe_package tree
+    Just "devhelp" -> return ()
+    Just format    -> fail ("The "++format++" format is not implemented")
 
 ppPrologue :: String -> Maybe Doc -> HtmlTable
 ppPrologue title Nothing = Html.emptyTable
@@ -215,45 +267,79 @@ ppPrologue title (Just doc) =
 ppModuleTree :: String -> [ModuleTree] -> HtmlTable
 ppModuleTree _ ts = 
   tda [theclass "section1"] << toHtml "Modules" </>
-  td <<  table ! [cellpadding 0, cellspacing 2] << 
-	(aboves (map (mkNode 0 []) ts) <-> mkPackages ts)
+  td << vanillaTable2 << htmlTable
+  where
+    genTable htmlTable id []     = (htmlTable,id)
+    genTable htmlTable id (x:xs) = genTable (htmlTable </> u) id' xs      
+      where
+        (u,id') = mkNode [] x 0 id
 
-mkNode :: Int -> [String] -> ModuleTree -> HtmlTable
-mkNode n ss (Node s leaf pkg []) =
-  mkLeaf n s ss leaf
-mkNode n ss (Node s leaf pkg ts) = 
-  mkLeaf n s ss leaf
-  </>
-  aboves (map (mkNode (n+1) (s:ss)) ts)
+    (htmlTable,_) = genTable emptyTable 0 ts
 
-mkLeaf :: Int -> String -> [String] -> Bool -> HtmlTable
-mkLeaf n s _ False = pad_td n << toHtml s
-mkLeaf n s ss True = pad_td n << ppHsModule mdl
-  where mdl = foldr (++) "" (s' : map ('.':) ss')
-	(s':ss') = reverse (s:ss)
+mkNode :: [String] -> ModuleTree -> Int -> Int -> (HtmlTable,Int)
+mkNode ss (Node s leaf pkg short ts) depth id = htmlNode
+  where
+    htmlNode = case ts of
+      [] -> (td_pad_w 1.25 depth << htmlModule  <-> shortDescr <-> htmlPkg,id)
+      _  -> (td_w depth << (collapsebutton id_s +++ htmlModule) <-> shortDescr <-> htmlPkg </> 
+                (td_subtree << sub_tree), id')
+
+    mod_width = 50::Int {-em-}
+
+    td_pad_w pad depth = 
+	tda [thestyle ("padding-left: " ++ show pad ++ "em;" ++
+		       "width: " ++ show (mod_width - depth*2) ++ "em")]
+
+    td_w depth = 
+	tda [thestyle ("width: " ++ show (mod_width - depth*2) ++ "em")]
+
+    td_subtree =
+	tda [thestyle ("padding: 0; padding-left: 2em")]
+
+    shortDescr :: HtmlTable
+    shortDescr = case short of
+	Nothing -> td empty
+	Just doc -> tda [theclass "rdoc"] (docToHtml doc)
+
+    htmlModule 
+      | leaf      = ppHsModule mdl
+      | otherwise = toHtml s
+
+    htmlPkg = case pkg of
+      Nothing -> td << empty
+      Just p  -> td << toHtml p
+
+    mdl = foldr (++) "" (s' : map ('.':) ss')
+    (s':ss') = reverse (s:ss)
 	 -- reconstruct the module name
-
-pad_td 0 = td
-pad_td n = tda [thestyle ("padding-left:" ++ show (n*20) ++ "px")]
-
-mkPackages :: [ModuleTree] -> HtmlTable
-mkPackages ts = aboves (map go ts)
-  where go (Node s leaf pkg ts) = tda [theclass "pkg"] << mkPkg pkg </> aboves (map go ts)
-
-mkPkg :: Maybe String -> Html
-mkPkg Nothing = empty
-mkPkg (Just p) = toHtml p
+    
+    id_s = "n:" ++ show id
+    
+    (sub_tree,id') = genSubTree emptyTable (id+1) ts
+    
+    genSubTree :: HtmlTable -> Int -> [ModuleTree] -> (Html,Int)
+    genSubTree htmlTable id [] = (sub_tree,id)
+      where
+        sub_tree = collapsed vanillaTable2 id_s htmlTable
+    genSubTree htmlTable id (x:xs) = genSubTree (htmlTable </> u) id' xs      
+      where
+        (u,id') = mkNode (s:ss) x (depth+1) id
 
 -- ---------------------------------------------------------------------------
 -- Generate the index
 
-ppHtmlIndex :: FilePath -> String -> Maybe String
-   -> [(Module,Interface)] -> IO ()
-ppHtmlIndex odir doctitle maybe_contents_url ifaces = do
+ppHtmlIndex :: FilePath
+            -> String 
+            -> Maybe String
+            -> Maybe String
+            -> Maybe String
+            -> [Interface] 
+            -> IO ()
+ppHtmlIndex odir doctitle maybe_package maybe_html_help_format maybe_contents_url ifaces = do
   let html = 
-	header (thetitle (toHtml (doctitle ++ " (Index)")) +++
-		thelink ! [href cssFile, 
-		  rel "stylesheet", thetype "text/css"]) +++
+	header (documentCharacterEncoding +++
+		thetitle (toHtml (doctitle ++ " (Index)")) +++
+		styleSheet) +++
         body << vanillaTable << (
 	    simpleHeader doctitle maybe_contents_url Nothing </>
 	    index_html
@@ -262,8 +348,15 @@ ppHtmlIndex odir doctitle maybe_contents_url ifaces = do
   when split_indices $
     mapM_ (do_sub_index index) initialChars
 
-  writeFile (odir ++ pathSeparator:indexHtmlFile) (renderHtml html)
-
+  writeFile (pathJoin [odir, indexHtmlFile]) (renderHtml html)
+  
+    -- Generate index and contents page for Html Help if requested
+  case maybe_html_help_format of
+    Nothing        -> return ()
+    Just "mshelp"  -> ppHHIndex  odir maybe_package ifaces
+    Just "mshelp2" -> ppHH2Index odir maybe_package ifaces
+    Just "devhelp" -> return ()
+    Just format    -> fail ("The "++format++" format is not implemented")
  where
   split_indices = length index > 50
 
@@ -281,16 +374,15 @@ ppHtmlIndex odir doctitle maybe_contents_url ifaces = do
 	    besides [ td << anchor ! [href (subIndexHtmlFile c)] <<
 			 toHtml [c]
 		    | c <- initialChars
-                    , any ((==c) . head . fst) index ]
+                    , any ((==c) . toUpper . head . fst) index ]
 
   do_sub_index this_ix c
     = unless (null index_part) $
-        writeFile (odir ++ pathSeparator:subIndexHtmlFile c)
-                  (renderHtml html)
+        writeFile (pathJoin [odir, subIndexHtmlFile c]) (renderHtml html)
     where 
-      html = header (thetitle (toHtml (doctitle ++ " (Index)")) +++
-		thelink ! [href cssFile, 
-		  rel "stylesheet", thetype "text/css"]) +++
+      html = header (documentCharacterEncoding +++
+		thetitle (toHtml (doctitle ++ " (Index)")) +++
+		styleSheet) +++
              body << vanillaTable << (
 	        simpleHeader doctitle maybe_contents_url Nothing </>
 		indexInitialLetterLinks </>
@@ -302,26 +394,27 @@ ppHtmlIndex odir doctitle maybe_contents_url ifaces = do
 
       index_part = [(n,stuff) | (n,stuff) <- this_ix, toUpper (head n) == c]
 
-  index :: [(String, FiniteMap HsQName [(Module,Bool)])]
-  index = sortBy cmp (fmToList full_index)
+  index :: [(String, Map HsQName [(Module,Bool)])]
+  index = sortBy cmp (Map.toAscList full_index)
     where cmp (n1,_) (n2,_) = n1 `compare` n2
 
   -- for each name (a plain string), we have a number of original HsNames that
   -- it can refer to, and for each of those we have a list of modules
   -- that export that entity.  Each of the modules exports the entity
   -- in a visible or invisible way (hence the Bool).
-  full_index :: FiniteMap String (FiniteMap HsQName [(Module,Bool)])
-  full_index = addListToFM_C (plusFM_C (++)) emptyFM
+  full_index :: Map String (Map HsQName [(Module,Bool)])
+  full_index = Map.fromListWith (flip (Map.unionWith (++)))
 		(concat (map getIfaceIndex ifaces))
 
-  getIfaceIndex (mdl,iface) = 
+  getIfaceIndex iface = 
     [ (hsNameStr nm, 
-	listToFM [(orig, [(mdl, not (nm `elemFM` iface_reexported iface))])])
-    | (nm, orig) <- fmToList (iface_env iface) ]
+	Map.fromList [(orig, [(mdl, not (nm `elem` iface_reexported iface))])])
+    | (nm, orig) <- Map.toAscList (iface_env iface) ]
+    where mdl = iface_module iface
 
-  indexElt :: (String, FiniteMap HsQName [(Module,Bool)]) -> HtmlTable
+  indexElt :: (String, Map HsQName [(Module,Bool)]) -> HtmlTable
   indexElt (str, entities) = 
-     case fmToList entities of
+     case Map.toAscList entities of
 	[(nm,entries)] ->  
 	    tda [ theclass "indexentry" ] << toHtml str <-> 
 			indexLinks (unQual nm) entries
@@ -343,7 +436,7 @@ ppHtmlIndex odir doctitle maybe_contents_url ifaces = do
   ppAnnot (HsVarName n)
        | isUpper c || c == ':'  = toHtml "Data Constructor"
        | otherwise		= toHtml "Function"
-      where c = head (ppHsIdentifier n)
+      where c = head (hsIdentifierStr n)
 
   indexLinks nm entries = 
      tda [ theclass "indexlinks" ] << 
@@ -361,20 +454,23 @@ ppHtmlIndex odir doctitle maybe_contents_url ifaces = do
 
 ppHtmlModule
 	:: FilePath -> String -> Maybe String -> Maybe String -> Maybe String
-	-> (Module,Interface) -> IO ()
+	-> Interface -> IO ()
 ppHtmlModule odir doctitle source_url 
-  maybe_contents_url maybe_index_url (Module mdl,iface) = do
-  let html = 
-	header (thetitle (toHtml mdl) +++
-		thelink ! [href cssFile,
-		  rel "stylesheet", thetype "text/css"]) +++
+  maybe_contents_url maybe_index_url iface = do
+  let 
+      Module mdl = iface_module iface
+      html = 
+	header (documentCharacterEncoding +++
+		thetitle (toHtml mdl) +++
+		styleSheet +++
+		(script ! [src jsFile, thetype "text/javascript"] $ noHtml)) +++
         body << vanillaTable << (
 	    pageHeader mdl iface doctitle source_url 
 		maybe_contents_url maybe_index_url </> s15 </>
 	    ifaceToHtml mdl iface </> s15 </>
 	    footer
          )
-  writeFile (moduleHtmlFile odir mdl) (renderHtml html)
+  writeFile (pathJoin [odir, moduleHtmlFile mdl]) (renderHtml html)
 
 ifaceToHtml :: String -> Interface -> HtmlTable
 ifaceToHtml _ iface 
@@ -400,6 +496,8 @@ ifaceToHtml _ iface
 
 	-- omit the synopsis if there are no documentation annotations at all
 	synopsis
+	  | no_doc_at_all = Html.emptyTable
+	  | otherwise
 	  = (tda [theclass "section1"] << toHtml "Synopsis") </>
 	    s15 </>
             (tda [theclass "body"] << vanillaTable <<
@@ -410,15 +508,12 @@ ifaceToHtml _ iface
 	-- if the documentation doesn't begin with a section header, then
 	-- add one ("Documentation").
 	maybe_doc_hdr
-	     | no_doc_at_all  = Html.emptyTable
-	     | otherwise = 
-		case exports of
+	    = case exports of		   
 		   [] -> Html.emptyTable
 		   ExportGroup _ _ _ : _ -> Html.emptyTable
 		   _ -> tda [ theclass "section1" ] << toHtml "Documentation"
 
-	bdy  | no_doc_at_all = []
-	     | otherwise     = map (processExport False) exports
+	bdy  = map (processExport False) exports
 
 ppModuleContents :: [ExportItem] -> HtmlTable
 ppModuleContents exports
@@ -434,7 +529,7 @@ ppModuleContents exports
     | lev <= n  = ( [], items )
     | otherwise = ( html:secs, rest2 )
     where
-	html = (dterm << linkedAnchor "" id0 << docToHtml doc)
+	html = (dterm << linkedAnchor id0 << docToHtml doc)
 		 +++ mk_subsections ssecs
 	(ssecs, rest1) = process lev rest
 	(secs,  rest2) = process n   rest1
@@ -578,13 +673,19 @@ ppHsDataDecl summary instances is_newty
 			 aboves (map ppSideBySideConstr cons)
 			)
 
+	inst_id = collapseId nm
+
 	instances_bit
 	   | null instances = Html.emptyTable
 	   | otherwise
-	   =  inst_hdr </>
-		 tda [theclass "body"] << spacedTable1 << (
-			aboves (map (declBox.ppInstHead) instances)
-		  )
+	   =  inst_hdr inst_id </>
+		 tda [theclass "body"] << 
+		    collapsed thediv inst_id (
+			spacedTable1 << (
+			   aboves (map (declBox.ppInstHead) instances)
+		        )
+ 		   )
+
 ppHsDataDecl _ _ _ _ d =
     error $ "HaddockHtml.ppHsDataDecl: unexpected decl " ++ show d
 
@@ -611,18 +712,18 @@ ppHsConstrHdr tvs ctxt
 
 ppSideBySideConstr :: HsConDecl -> HtmlTable
 ppSideBySideConstr (HsConDecl _ nm tvs ctxt typeList doc) =
-  declBox (hsep ((ppHsConstrHdr tvs ctxt +++ 
+  argBox (hsep ((ppHsConstrHdr tvs ctxt +++ 
 		ppHsBinder False nm) : map ppHsBangType typeList)) <->
   maybeRDocBox doc
 ppSideBySideConstr (HsRecDecl _ nm tvs ctxt fields doc) =
-  declBox (ppHsConstrHdr tvs ctxt +++ ppHsBinder False nm) <->
+  argBox (ppHsConstrHdr tvs ctxt +++ ppHsBinder False nm) <->
   maybeRDocBox doc </>
   (tda [theclass "body"] << spacedTable1 <<
      aboves (map ppSideBySideField fields))
 
 ppSideBySideField :: HsFieldDecl -> HtmlTable
 ppSideBySideField (HsFieldDecl ns ty doc) =
-  declBox (hsep (punctuate comma (map (ppHsBinder False) ns))
+  argBox (hsep (punctuate comma (map (ppHsBinder False) ns))
 	   <+> dcolon <+> ppHsBangType ty) <->
   maybeRDocBox doc
 
@@ -752,11 +853,16 @@ ppHsClassDecl summary instances orig_c
 			             ]
 			)
 
-	instances_bit
-	   =  s8 </> inst_hdr </>
-		 tda [theclass "body"] << spacedTable1 << (
+	inst_id = collapseId nm
+ 	instances_bit
+	   | null instances = Html.emptyTable
+	   | otherwise 
+	   =  s8 </> inst_hdr inst_id </>
+		 tda [theclass "body"] << 
+		   collapsed thediv inst_id (
+		      spacedTable1 << (
 			aboves (map (declBox.ppInstHead) instances)
-		  )
+		  ))
 
 ppHsClassDecl _ _ _ d =
     error $ "HaddockHtml.ppHsClassDecl: unexpected decl: " ++ show d
@@ -791,26 +897,26 @@ ppFunSig summary nm ty0 doc
 
 	do_args :: Html -> HsType -> HtmlTable
 	do_args leader (HsForAllType (Just tvs) ctxt ty)
-	  = (declBox (
+	  = (argBox (
 		leader <+> 
 		hsep (keyword "forall" : map ppHsName tvs ++ [toHtml "."]) <+>
 		ppHsIPContext ctxt)
 	      <-> rdocBox noHtml) </> 
 	    do_args darrow ty
 	do_args leader (HsForAllType Nothing ctxt ty)
-	  = (declBox (leader <+> ppHsIPContext ctxt)
+	  = (argBox (leader <+> ppHsIPContext ctxt)
 		<-> rdocBox noHtml) </> 
 	    do_args darrow ty
 	do_args leader (HsTyFun (HsTyDoc ty doc0) r)
-	  = (declBox (leader <+> ppHsBType ty) <-> rdocBox (docToHtml doc0))
+	  = (argBox (leader <+> ppHsBType ty) <-> rdocBox (docToHtml doc0))
             </> do_args arrow r
 	do_args leader (HsTyFun ty r)
-	  = (declBox (leader <+> ppHsBType ty) <-> rdocBox noHtml) </>
+	  = (argBox (leader <+> ppHsBType ty) <-> rdocBox noHtml) </>
 	    do_args arrow r
 	do_args leader (HsTyDoc ty doc0)
-	  = (declBox (leader <+> ppHsBType ty) <-> rdocBox (docToHtml doc0))
+	  = (argBox (leader <+> ppHsBType ty) <-> rdocBox (docToHtml doc0))
 	do_args leader ty
-	  = declBox (leader <+> ppHsBType ty) <-> rdocBox (noHtml)
+	  = argBox (leader <+> ppHsBType ty) <-> rdocBox (noHtml)
 
 -- ----------------------------------------------------------------------------
 -- Types and contexts
@@ -887,23 +993,10 @@ isSpecial _                                      = False
 ppHsName :: HsName -> Html
 ppHsName nm = toHtml (hsNameStr nm)
 
-hsAnchorNameStr :: HsName -> String
-hsAnchorNameStr (HsTyClsName id0) = "t:" ++ ppHsIdentifier id0
-hsAnchorNameStr (HsVarName id0)   = "v:" ++ ppHsIdentifier id0
-
-hsNameStr :: HsName -> String
-hsNameStr (HsTyClsName id0) = ppHsIdentifier id0
-hsNameStr (HsVarName id0)   = ppHsIdentifier id0
-
-ppHsIdentifier :: HsIdentifier -> String
-ppHsIdentifier (HsIdent str)   =  str
-ppHsIdentifier (HsSymbol str)  =  str
-ppHsIdentifier (HsSpecial str) =  str
-
 ppHsBinder :: Bool -> HsName -> Html
 -- The Bool indicates whether we are generating the summary, in which case
 -- the binder will be a link to the full definition.
-ppHsBinder True nm = linkedAnchor "" (hsAnchorNameStr nm) << ppHsBinder' nm
+ppHsBinder True nm = linkedAnchor (hsAnchorNameStr nm) << ppHsBinder' nm
 ppHsBinder False nm = linkTarget nm +++ bold << ppHsBinder' nm
 
 ppHsBinder' :: HsName -> Html
@@ -916,19 +1009,14 @@ ppHsBindIdent (HsSymbol str)  =  parens (toHtml str)
 ppHsBindIdent (HsSpecial str) =  toHtml str
 
 linkId :: Module -> Maybe HsName -> Html -> Html
-linkId (Module mdl) mbStr = linkedAnchor (moduleHtmlFile fp mdl) frag
-  where frag = case mbStr of
-                  Nothing  -> ""
-                  Just str -> hsAnchorNameStr str
-        fp   = case lookupFM html_xrefs (Module mdl) of
-		  Nothing  -> ""
-		  Just fp0 -> fp0
+linkId (Module mdl) mbName = anchor ! [href hr]
+  where hr = case mbName of
+                  Nothing   -> moduleHtmlFile mdl
+                  Just name -> nameHtmlRef mdl name
 
 ppHsModule :: String -> Html
-ppHsModule mdl = anchor ! [href ((moduleHtmlFile fp modname) ++ ref)] << toHtml mdl
-  where fp = case lookupFM html_xrefs (Module modname) of
-		Just fp0 -> fp0 
-		Nothing  -> ""
+ppHsModule mdl = anchor ! [href ((moduleHtmlFile modname) ++ ref)] << toHtml mdl
+  where 
         (modname,ref) = break (== '#') mdl
 
 -- -----------------------------------------------------------------------------
@@ -1036,6 +1124,12 @@ text   = strAttr "TEXT"
 declBox :: Html -> HtmlTable
 declBox html = tda [theclass "decl"] << html
 
+-- a box for displaying an 'argument' (some code which has text to the
+-- right of it).  Wrapping is not allowed in these boxes, whereas it is
+-- in a declBox.
+argBox :: Html -> HtmlTable
+argBox html = tda [theclass "arg"] << html
+
 -- a box for displaying documentation, 
 -- indented and with a little padding at the top
 docBox :: Html -> HtmlTable
@@ -1061,16 +1155,20 @@ topButBox html = tda [theclass "topbut"] << html
 -- a narrow table is the same but without width 100%.
 vanillaTable, narrowTable :: Html -> Html
 vanillaTable = table ! [theclass "vanilla", cellspacing 0, cellpadding 0]
+vanillaTable2 = table ! [theclass "vanilla2", cellspacing 0, cellpadding 0]
 narrowTable  = table ! [theclass "narrow",  cellspacing 0, cellpadding 0]
 
 spacedTable1, spacedTable5 :: Html -> Html
 spacedTable1 = table ! [theclass "vanilla",  cellspacing 1, cellpadding 0]
 spacedTable5 = table ! [theclass "vanilla",  cellspacing 5, cellpadding 0]
 
-constr_hdr, meth_hdr, inst_hdr :: HtmlTable
-constr_hdr = tda [ theclass "section4" ] << toHtml "Constructors"
-meth_hdr   = tda [ theclass "section4" ] << toHtml "Methods"
-inst_hdr   = tda [ theclass "section4" ] << toHtml "Instances"
+constr_hdr, meth_hdr :: HtmlTable
+constr_hdr  = tda [ theclass "section4" ] << toHtml "Constructors"
+meth_hdr    = tda [ theclass "section4" ] << toHtml "Methods"
+
+inst_hdr :: String -> HtmlTable
+inst_hdr id = 
+  tda [ theclass "section4" ] << (collapsebutton id +++ toHtml " Instances")
 
 dcolon, arrow, darrow :: Html
 dcolon = toHtml "::"
@@ -1084,10 +1182,36 @@ s15 = tda [ theclass "s15" ] << noHtml
 namedAnchor :: String -> Html -> Html
 namedAnchor n = anchor ! [name (escapeStr n)]
 
-linkedAnchor :: String -> String -> Html -> Html
-linkedAnchor ref frag = anchor ! [href hr]
-   where hr | null frag = ref
-            | otherwise = ref ++ '#': escapeStr frag
+--
+-- A section of HTML which is collapsible via a +/- button.
+--
 
-escapeStr :: String -> String
-escapeStr = flip escapeString unreserved
+-- TODO: Currently the initial state is non-collapsed. Change the 'minusFile'
+-- below to a 'plusFile' and the 'display:block;' to a 'display:none;' when we
+-- use cookies from JavaScript to have a more persistent state.
+
+collapsebutton :: String -> Html
+collapsebutton id = 
+  image ! [ src minusFile, theclass "coll", onclick ("toggle(this,'" ++ id ++ "')"), alt "show/hide" ]
+
+collapsed :: (HTML a) => (Html -> Html) -> String -> a -> Html
+collapsed fn id html =
+  fn ! [identifier id, thestyle "display:block;"] << html
+
+-- A quote is a valid part of a Haskell identifier, but it would interfere with
+-- the ECMA script string delimiter used in collapsebutton above.
+collapseId :: HsName -> String
+collapseId nm = "i:" ++ escapeStr (hsNameStr nm)
+
+linkedAnchor :: String -> Html -> Html
+linkedAnchor frag = anchor ! [href hr]
+   where hr | null frag = ""
+            | otherwise = '#': escapeStr frag
+
+documentCharacterEncoding :: Html
+documentCharacterEncoding =
+   meta ! [httpequiv "Content-Type", content "text/html; charset=UTF-8"]
+
+styleSheet :: Html
+styleSheet =
+   thelink ! [href cssFile, rel "stylesheet", thetype "text/css"]
